import logging

from vortex.handler.TupleActionProcessorProxy import TupleActionProcessorProxy
from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)

from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_base.client.PluginClientEntryHookABC import (
    PluginClientEntryHookABC,
)
from peek_plugin_eventdb._private.PluginNames import (
    eventdbFilt,
    eventdbObservableName,
    eventdbActionProcessorName,
)
from peek_plugin_eventdb._private.storage.DeclarativeBase import (
    loadStorageTuples,
)
from peek_plugin_eventdb._private.tuples import loadPrivateTuples
from peek_plugin_eventdb.tuples import loadPublicTuples

logger = logging.getLogger(__name__)


class ClientEntryHook(PluginClientEntryHookABC):
    def __init__(self, *args, **kwargs):
        PluginClientEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

    def load(self) -> None:
        loadStorageTuples()
        loadPrivateTuples()
        loadPublicTuples()

        logger.debug("Loaded")

    def start(self):
        # ----------------
        # Proxy actions back to the server, we don't process them at all
        self._loadedObjects.append(
            TupleActionProcessorProxy(
                tupleActionProcessorName=eventdbActionProcessorName,
                proxyToVortexName=peekServerName,
                additionalFilt=eventdbFilt,
            )
        )

        # ----------------
        # Provide the devices access to the servers observable
        tupleDataObservableProxyHandler = TupleDataObservableProxyHandler(
            observableName=eventdbObservableName,
            proxyToVortexName=peekServerName,
            additionalFilt=eventdbFilt,
            observerName="Proxy to devices",
        )
        self._loadedObjects.append(tupleDataObservableProxyHandler)

        # Support file downloads for device updates
        # noinspection PyTypeChecker
        proxyResource = self.createProxy()

        # Matches resource path on server
        # noinspection PyTypeChecker
        self.platform.addOfficeResource(b"download", proxyResource)

        logger.debug("Started")

    def stop(self):
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        logger.debug("Stopped")

    def unload(self):
        logger.debug("Unloaded")



from vortex.handler.TupleActionProcessor import TupleActionProcessor

from peek_plugin_eventdb._private.PluginNames import eventdbFilt
from peek_plugin_eventdb._private.PluginNames import eventdbActionProcessorName
from .controller.MainController import MainController


def makeTupleActionProcessorHandler(mainController: MainController):
    processor = TupleActionProcessor(
        tupleActionProcessorName=eventdbActionProcessorName,
        additionalFilt=eventdbFilt,
        defaultDelegate=mainController,
    )
    return processor



from sqlalchemy import Column, ForeignKey, Boolean, Index
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship
from vortex.Tuple import addTupleType, Tuple, TupleField

from peek_plugin_eventdb._private.PluginNames import eventdbTuplePrefix
from .DeclarativeBase import DeclarativeBase
from .EventDBModelSetTable import EventDBModelSetTable
from ...tuples.EventDBPropertyTuple import EventDBPropertyTuple


@addTupleType
class EventDBPropertyTable(Tuple, DeclarativeBase):
    __tablename__ = "EventDBProperty"
    __tupleType__ = eventdbTuplePrefix + "EventDBPropertyTable"

    SHOW_FILTER_AS_FREE_TEXT = 1
    SHOW_FILTER_SELECT_MANY = 2
    SHOW_FILTER_SELECT_ONE = 3

    id = Column(Integer, primary_key=True, autoincrement=True)

    modelSetId = Column(
        Integer, ForeignKey("EventDBModelSet.id", ondelete="CASCADE"), nullable=False
    )
    modelSet = relationship(EventDBModelSetTable)

    key = Column(String, nullable=False)
    name = Column(String, nullable=False)
    order = Column(Integer, nullable=False)
    comment = Column(String)

    useForFilter = Column(Boolean)
    useForDisplay = Column(Boolean)
    useForPopup = Column(Boolean)
    showFilterAs = Column(Integer)

    displayByDefaultOnSummaryView = Column(Boolean)
    displayByDefaultOnDetailView = Column(Boolean)

    valuesFromAdminUi = TupleField()

    __table_args__ = (
        Index("idx_EventDBProp_name", modelSetId, key, unique=True),
        Index("idx_EventDBProp_value", modelSetId, name, unique=True),
    )

    def toTuple(self) -> EventDBPropertyTuple:
        return EventDBPropertyTuple(
            modelSetKey=self.modelSet.key,
            key=self.key,
            name=self.name,
            order=self.order,
            comment=self.comment,
            useForFilter=self.useForFilter,
            useForDisplay=self.useForDisplay,
            useForPopup=self.useForPopup,
            displayByDefaultOnSummaryView=self.displayByDefaultOnSummaryView,
            displayByDefaultOnDetailView=self.displayByDefaultOnDetailView,
            showFilterAs=self.showFilterAs,
            values=[v.toTuple() for v in self.values],
        )



from datetime import datetime

from vortex.Tuple import addTupleType, TupleField, Tuple

from peek_plugin_eventdb._private.PluginNames import eventdbTuplePrefix


@addTupleType
class AdminStatusTuple(Tuple):
    __tupleType__ = eventdbTuplePrefix + "AdminStatusTuple"

    addedEvents: int = TupleField(0)
    removedEvents: int = TupleField(0)
    updatedAlarmFlags: int = TupleField(0)
    lastActivity: datetime = TupleField()



from typing import Optional

from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_plugin_eventdb._private.PluginNames import eventdbTuplePrefix


@addTupleType
class EventDBPropertyValueTuple(Tuple):
    """Event DB Property Value Tuple

    This tuple is an option that the user can select for each property.

    For example:

    Property = Alarm Class
    Property Value = Class 1
    Property Value = Class 2

    name: The name of this property value that is displayed to the user.

    value: The value that matches
       EvenDBEventTuple.value[property.key] == propertyValue.value

    color: If this property has a color then define it here.
           the first property value with a color will be applied
           ordered by property.order

    comment: The tooltip to display to the user in a UI.

    """

    __tupleType__ = eventdbTuplePrefix + "EventDBPropertyValueTuple"

    name: str = TupleField()
    value: str = TupleField()
    color: str = TupleField()
    comment: Optional[str] = TupleField()



from datetime import datetime
from typing import List

from peek_plugin_eventdb.tuples.EventDBEventTuple import EventDBEventTuple
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleSelectorUpdateMapperABC


class NewEventsTupleSelector(TupleSelector):
    """New Events Tuple Selector

    This is a tuple selector that is only used to notify our mapper below that new
    events have been imported.

    The mapper then maps to the relevent real tuple selectors
    """

    NAME = "new events imported"

    def __init__(self, minDate: datetime, maxDate: datetime):
        TupleSelector.__init__(self, self.NAME, dict(minDate=minDate, maxDate=maxDate))

    @property
    def minDate(self):
        return self.selector["minDate"]

    @property
    def maxDate(self):
        return self.selector["maxDate"]


class NewEventTSUpdateMapper(TupleSelectorUpdateMapperABC):
    def mapTupleSelector(
        self,
        triggerTupleSelector: TupleSelector,
        allTupleSelectors: List[TupleSelector],
    ) -> List[TupleSelector]:
        if not triggerTupleSelector.name == NewEventsTupleSelector.NAME:
            return []

        eventTss = filter(
            lambda ts: ts.name == EventDBEventTuple.tupleName(), allTupleSelectors
        )

        updateTs: NewEventsTupleSelector = triggerTupleSelector
        minDate = updateTs.minDate
        maxDate = updateTs.maxDate

        results = []
        for tupleSelector in eventTss:
            newestDateTime = tupleSelector.selector.get("newestDateTime")
            oldestDateTime = tupleSelector.selector.get("oldestDateTime")

            isMinInRange = (oldestDateTime is None or oldestDateTime <= minDate) and (
                newestDateTime is None or minDate <= newestDateTime
            )

            isMaxInRange = (oldestDateTime is None or oldestDateTime <= maxDate) and (
                newestDateTime is None or maxDate <= newestDateTime
            )

            if isMinInRange or isMaxInRange:
                results.append(tupleSelector)

        return results



import json
import logging
from typing import Dict

from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from txhttputil.site.BasicResource import BasicResource
from vortex.TupleSelector import TupleSelector

from peek_plugin_base.storage.DbConnection import DbSessionCreator
from peek_plugin_base.storage.RunPyInPg import runPyInPg
from peek_plugin_eventdb._private.server.tuple_providers.EventDBEventTupleProvider import (
    EventDBEventTupleProvider,
)

logger = logging.getLogger(__name__)


class _Prop:
    def __init__(self, name: str):
        self.name = name
        self.nameByValueMap = {}


class DownloadEventsResource(BasicResource):
    def __init__(self, dbSessionCreator: DbSessionCreator):
        Resource.__init__(self)
        self._dbSessionCreator = dbSessionCreator

    def render_GET(self, request):
        tupleSelectorStr = request.args[b"tupleSelector"][0].decode()

        request.responseHeaders.setRawHeaders("content-type", ["text/csv"])
        logger.info("Received download request for events.")
        logger.debug(
            "Download request received with TupleSelector = %s.", tupleSelectorStr
        )

        def good(data):
            request.write(data.encode())
            request.finish()
            logger.info("Finished download of events")

        def bad(failure):
            request.setResponseCode(500)
            e = failure.value
            logger.exception(e)

            request.finish()

        d = runPyInPg(
            logger,
            self._dbSessionCreator,
            self._loadInPg,
            EventDBEventTupleProvider._importTuples,
            tupleSelectorStr=tupleSelectorStr,
        )

        d.addCallbacks(good, bad)

        def closedError(failure):
            logger.error("Got closedError %s" % failure)

        request.notifyFinish().addErrback(closedError)

        return NOT_DONE_YET

    @classmethod
    def _loadInPg(cls, plpy, tupleSelectorStr: str):
        tupleSelector = TupleSelector.fromJsonStr(tupleSelectorStr)
        columnPropKeys = tupleSelector.selector["columnPropKeys"]

        modelSetId = EventDBEventTupleProvider.getModelSetId(
            plpy, tupleSelector.selector["modelSetKey"]
        )

        propByKey = cls._getPropertyByKey(plpy, modelSetId)

        tuples = EventDBEventTupleProvider.loadTuples(plpy, tupleSelector)

        data = ""

        # Write the heading
        cols = ["Date", "Time", "Milliseconds", "UTC Offset"]
        for columnPropKey in columnPropKeys:
            cols.append('"%s"' % propByKey[columnPropKey].name)

        data += ",".join(cols) + "\r\n"

        # Write the data of the table
        for row in tuples:
            value = json.loads(row.value)
            tz = str(row.dateTime.strftime("%z"))
            cols = [
                row.dateTime.strftime("%d-%b-%Y"),
                row.dateTime.strftime("%H:%M:%S"),
                str(int(row.dateTime.microsecond / 1000)),
                "%s:%s" % (tz[:-2], tz[-2:]),
            ]

            for columnPropKey in columnPropKeys:
                prop = propByKey[columnPropKey]
                val = value.get(columnPropKey)

                if val is None:
                    cols.append("")
                else:
                    # Map the value if it has one.
                    valKey = str(val).replace("null", "none").lower()
                    val = prop.nameByValueMap.get(valKey, val)

                    if isinstance(val, int):
                        cols.append(str(val))
                    else:
                        cols.append('"%s"' % val)

            data += ",".join(cols) + "\r\n"

        return data

    @classmethod
    def _getPropertyByKey(cls, plpy, modelSetId) -> Dict[str, _Prop]:
        # Load in the ModelSet ID
        sql = (
            """
              SELECT id, key, name
              FROM pl_eventdb."EventDBProperty"
              where "modelSetId" = %s
              """
            % modelSetId
        )
        rows = plpy.execute(sql, 10000)

        if not len(rows):
            raise Exception("No properties exist with modelSetId %s" % modelSetId)

        propByKey = {}
        propById = {}
        for row in rows:
            prop = _Prop(row["name"])
            propByKey[row["key"]] = prop
            propById[row["id"]] = prop

        # Load in the Property Values

        sql = (
            """
              SELECT v."propertyId", v.name, v.value
              FROM pl_eventdb."EventDBPropertyValue" as v
                  JOIN pl_eventdb."EventDBProperty" as p on v."propertyId"=p.id
              where "modelSetId" = %s
              """
            % modelSetId
        )
        rows = plpy.execute(sql, 10000)

        for row in rows:
            prop = propById[row["propertyId"]]
            key = str(row["value"]).replace("null", "none").lower()
            prop.nameByValueMap[key] = row["name"]

        return propByKey



from abc import ABCMeta, abstractmethod
from typing import List

from rx.subjects import Subject


class EventDBReadApiABC(metaclass=ABCMeta):
    @abstractmethod
    def newEventsObservable(self, modelSetKey: str) -> Subject:
        """Raw Value Update Observable

        Return an observable that fires with lists of C{EventDBEventTuple} tuples
        containing updates to EventDB values.

        :param modelSetKey:  The name of the model set for the EventDB

        :return: An observable that fires when values are updated in the eventdb
        :rtype: Subject[List[EventDBEventTuple]]

        """



from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.sql.expression import cast, null, case
from sqlalchemy.sql.schema import Index
from sqlalchemy.types import Integer, String, Boolean
from vortex.Tuple import Tuple, addTupleType

from peek_plugin_eventdb._private.PluginNames import eventdbTuplePrefix
from .DeclarativeBase import DeclarativeBase

"""Mapping a polymorphic-valued vertical table as a dictionary.

For any given properties row, the value of the 'type' column will point to the
'_value' column active for that row.

It only differs in the mapping for vertical rows.  Here,
we'll use a @hybrid_property to attune a smart '.value' attribute that wraps up
reading and writing those various '_value' columns and keeps the '.type' up to
date.

"""

from sqlalchemy.orm.interfaces import PropComparator
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy import event
from sqlalchemy import literal_column


class ProxiedDictMixin(object):
    """Adds obj[key] access to a mapped class.

    This class basically proxies dictionary access to an attribute
    called ``_proxied``.  The class which inherits this class
    should have an attribute called ``_proxied`` which points to a dictionary.

    """

    _proxied = None

    def __len__(self):
        return len(self._proxied)

    def __iter__(self):
        return iter(self._proxied)

    def __getitem__(self, key):
        return self._proxied[str(key)]

    def __contains__(self, key):
        return str(key) in self._proxied

    def __setitem__(self, key, value):
        self._proxied[str(key)] = value

    def __delitem__(self, key):
        del self._proxied[str(key)]


class PolymorphicVerticalProperty(object):
    """A key/value pair with polymorphic value storage.

    The class which is mapped should indicate typing information
    within the "info" dictionary of mapped Column objects; see
    the AnimalFact mapping below for an example.

    """

    def __init__(self, key, value=None):
        self.key = key
        self.value = value

    @hybrid_property
    def value(self):
        fieldname, discriminator = self.type_map[self.type]
        if fieldname is None:
            return None
        else:
            return getattr(self, fieldname)

    @value.setter
    def value(self, value):
        py_type = type(value)
        fieldname, discriminator = self.type_map[py_type]

        self.type = discriminator
        if fieldname is not None:
            setattr(self, fieldname, value)

    @value.deleter
    def value(self):
        self._set_value(None)

    @value.comparator
    class value(PropComparator):
        """A comparator for .value, builds a polymorphic comparison via CASE."""

        def __init__(self, cls):
            self.cls = cls

        def _case(self):
            pairs = set(self.cls.type_map.values())
            whens = [
                (
                    literal_column("'%s'" % discriminator),
                    cast(getattr(self.cls, attribute), String),
                )
                for attribute, discriminator in pairs
                if attribute is not None
            ]
            return case(whens, self.cls.type, null())

        def __eq__(self, other):
            return self._case() == cast(other, String)

        def __ne__(self, other):
            return self._case() != cast(other, String)

    def __repr__(self):
        return "<%s %r=%r>" % (self.__class__.__name__, self.key, self.value)


@addTupleType
class SettingProperty(PolymorphicVerticalProperty, Tuple, DeclarativeBase):
    """A setting property."""

    __tablename__ = "SettingProperty"
    __tupleType__ = eventdbTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)
    settingId = Column(ForeignKey("Setting.id"), nullable=False)
    key = Column(String(50), nullable=False)
    type = Column(String(16))

    # add information about storage for different types
    # in the info dictionary of Columns
    int_value = Column(Integer, info={"type": (int, "integer")})
    char_value = Column(String(50), info={"type": (str, "string")})
    boolean_value = Column(Boolean, info={"type": (bool, "boolean")})

    def __init__(self, key=None, value=None):
        PolymorphicVerticalProperty.__init__(self, key=key, value=value)
        Tuple.__init__(self)

    __table_args__ = (Index("idx_SettingProperty_settingId", settingId),)


@addTupleType
class Setting(ProxiedDictMixin, Tuple, DeclarativeBase):
    """an Animal"""

    __tablename__ = "Setting"
    __tupleType__ = eventdbTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(50))

    properties = relationship(
        "SettingProperty", collection_class=attribute_mapped_collection("key")
    )

    propertyObjects = relationship("SettingProperty", viewonly=True, lazy=False)

    _proxied = association_proxy(
        "properties",
        "value",
        creator=lambda key, value: SettingProperty(key=key, value=value),
    )

    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        return "Setting(%r)" % self.name

    @classmethod
    def with_characteristic(self, key, value):
        return self.properties.any(key=key, value=value)


@event.listens_for(SettingProperty, "mapper_configured", propagate=True)
def on_new_class(mapper, cls_):
    """Look for Column objects with type info in them, and work up
    a lookup table."""

    info_dict = {}
    info_dict[type(None)] = (None, "none")
    info_dict["none"] = (None, "none")

    for k in list(mapper.c.keys()):
        col = mapper.c[k]
        if "type" in col.info:
            python_type, discriminator = col.info["type"]
            info_dict[python_type] = (k, discriminator)
            info_dict[discriminator] = (k, discriminator)
    cls_.type_map = info_dict


class PropertyKey(object):
    def __init__(self, name, defaultValue, propertyDict):
        self.name = name
        self.defaultValue = defaultValue
        propertyDict[name] = self

    def __repr__(self):
        return self.name


def _getSetting(ormSession, name, propertyDict, key=None, value=None):
    all = ormSession.query(Setting).filter(Setting.name == name).all()

    needsCommit = False

    if all:
        setting = all[0]
        ormSession.expire(setting)
    else:
        setting = Setting(name)
        ormSession.add(setting)
        needsCommit = True

    for prop in list(propertyDict.values()):
        if not prop.name in setting:
            setting[prop.name] = prop.defaultValue
            needsCommit = True

    if needsCommit:
        ormSession.commit()

    if not key:
        return setting

    # Make sure the propery is defined for this setting
    assert str(key) in propertyDict, "Key %s is not defined in setting %s" % (key, name)

    if value is None:
        return setting[key]

    setting[key] = value
    ormSession.commit()

    # Close after return
    return setting[key]


# =============================================================================
# GLOBAL PROPERTIES
# =============================================================================

globalProperties = {}


def globalSetting(ormSession, key=None, value=None):
    return _getSetting(ormSession, "Global", globalProperties, key=key, value=value)


#
# VALUE_UPDATER_ENABLED = PropertyKey('Value Updater Enabled', True,
#                                     propertyDict=globalProperties)



from sqlalchemy import Column, ForeignKey, Index
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship
from vortex.Tuple import addTupleType, Tuple

from peek_plugin_eventdb._private.PluginNames import eventdbTuplePrefix
from .DeclarativeBase import DeclarativeBase
from .EventDBPropertyTable import EventDBPropertyTable
from ...tuples.EventDBPropertyValueTuple import EventDBPropertyValueTuple


@addTupleType
class EventDBPropertyValueTable(Tuple, DeclarativeBase):
    __tablename__ = "EventDBPropertyValue"
    __tupleType__ = eventdbTuplePrefix + "EventDBPropertyValueTable"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    value = Column(String, nullable=False)
    color = Column(String)
    comment = Column(String)

    propertyId = Column(
        Integer, ForeignKey("EventDBProperty.id", ondelete="CASCADE"), nullable=False
    )
    property = relationship(EventDBPropertyTable, backref="values")

    __table_args__ = (
        Index("idx_EventDBPropVal_name", propertyId, name, unique=True),
        Index("idx_EventDBPropVal_value", propertyId, value, unique=True),
    )

    def toTuple(self) -> EventDBPropertyValueTuple:
        return EventDBPropertyValueTuple(
            name=self.name, value=self.value, color=self.color, comment=self.comment
        )



from abc import ABCMeta, abstractmethod

from peek_plugin_eventdb.server.EventDBReadApiABC import EventDBReadApiABC
from peek_plugin_eventdb.server.EventDBWriteApiABC import EventDBWriteApiABC


class EventDBApiABC(metaclass=ABCMeta):
    @property
    @abstractmethod
    def writeApi(self) -> EventDBWriteApiABC:
        """EventDB Write API

        This API is for all the methods that make changes to the EventDB

        :return: A reference to the EventDBWriteApiABC instance

        """

    @property
    @abstractmethod
    def readApi(self) -> EventDBReadApiABC:
        """EventDB Read API

        This API is for all the methods to read changes from the EventDB

        :return: A reference to the EventDBReadApiABC instance

        """



from abc import ABCMeta, abstractmethod
from typing import List

from twisted.internet.defer import Deferred


class EventDBWriteApiABC(metaclass=ABCMeta):
    @abstractmethod
    def addEvents(self, modelSetKey: str, eventsEncodedPayload: str) -> Deferred:
        """Add Events

        Add events to the EventDB

        :param modelSetKey:  The name of the model set for the EventDB
        :param eventsEncodedPayload: An encoded Payload containing a
         list of events to insert.

        :return: A deferred that fires when the insert is complete.
        :rtype: None

        """

    @abstractmethod
    def removeEvents(self, modelSetKey: str, eventKeys: List[str]) -> Deferred:
        """Remove Events

        Remove events from the EventDB

        :param modelSetKey:  The name of the model set for the EventDB
        :param eventKeys: An list of event keys to remove.

        :return: A deferred that fires when the removal is complete.
        :rtype: None

        """

    @abstractmethod
    def updateAlarmFlags(
        self, modelSetKey: str, eventKeys: List[str], alarmFlag: bool
    ) -> Deferred:
        """Update Alarm Flag

        Change the value of the alarm flag for alarms.

        :param modelSetKey:  The name of the model set for the EventDB
        :param eventKeys: An list of event keys to update the alarm flag.
        :param alarmFlag: The value to set the alarm flag to.

        :return: A deferred that fires when the update is complete.
        :rtype: None

        """

    @abstractmethod
    def replaceProperties(
        self, modelSetKey: str, propertiesEncodedPayload: str
    ) -> Deferred:
        """Replace Properties

        Create or Replace the properties for a model set of events.

        :param modelSetKey:  The name of the model set for the EventDB
        :param propertiesEncodedPayload: An encoded Payload containing a
                payload.tuples=List[EventDBPropertyTuple]
         list of events to insert.

        :return: A deferred that fires when the create / update is complete.
        :rtype: None

        """



import logging

from sqlalchemy import Column, DateTime, BigInteger, String, Integer, Boolean
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.sql.schema import Index, ForeignKey
from vortex.Tuple import Tuple, addTupleType

from peek_plugin_eventdb._private.PluginNames import eventdbTuplePrefix
from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class EventDBEventTable(Tuple, DeclarativeBase):
    __tablename__ = "EventDBEvent"
    __tupleType__ = eventdbTuplePrefix + "EventDBEventTable"

    id = Column(BigInteger, primary_key=True, autoincrement=True)

    dateTime = Column(DateTime(True), primary_key=True, nullable=False)

    value = Column(JSONB, nullable=False)

    key = Column(String)

    isAlarm = Column(Boolean)

    modelSetId = Column(
        Integer, ForeignKey("EventDBModelSet.id", ondelete="CASCADE"), nullable=False
    )

    __table_args__ = (
        Index("idx_EventDBEvent_modelSetId_key", modelSetId, key, unique=False),
    )



import logging
import os
from datetime import datetime
from typing import List
from typing import Optional

import pytz
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.task import LoopingCall
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_plugin_base.storage.DbConnection import DbSessionCreator
from peek_plugin_base.storage.RunPyInPg import runPyInPg
from peek_plugin_base.util.PeekPsUtil import PeekPsUtil
from peek_plugin_eventdb._private.server.EventDBReadApi import EventDBReadApi
from peek_plugin_eventdb._private.server.controller.AdminStatusController import (
    AdminStatusController,
)
from peek_plugin_eventdb._private.server.controller.EventDBImportEventsInPgTask import (
    EventDBImportEventsInPgTask,
)
from peek_plugin_eventdb._private.server.controller.EventDBImportPropertiesInPgTask import (
    EventDBImportPropertiesInPgTask,
)
from peek_plugin_eventdb._private.server.tuple_selector_mappers.NewEventTSUpdateMapper import (
    NewEventsTupleSelector,
)
from peek_plugin_eventdb._private.storage.EventDBPropertyTable import (
    EventDBPropertyTable,
)
from peek_plugin_eventdb.tuples.EventDBPropertyTuple import EventDBPropertyTuple

logger = logging.getLogger(__name__)


class EventDBImportController:
    """EventDB Import Controller"""

    MAX_CPU_PERCENTAGE = 50.0
    NOTIFY_TIME_SECONDS = 5.0

    def __init__(
        self,
        dbSessionCreator: DbSessionCreator,
        statusController: AdminStatusController,
        tupleObservable: TupleDataObservableHandler,
    ):
        self._dbSessionCreator = dbSessionCreator
        self._statusController = statusController
        self._tupleObservable = tupleObservable

        self._updateSelectorQueue: list[NewEventsTupleSelector] = []
        self._updateNotifyLoopingCall: Optional[LoopingCall] = None

        self._process = None

    def start(self):
        assert not self._updateNotifyLoopingCall, "We've been started already"
        self._updateNotifyLoopingCall = LoopingCall(self._batchNotifyUpdates)
        self._updateNotifyLoopingCall.start(self.NOTIFY_TIME_SECONDS)

    def setReadApi(self, readApi: EventDBReadApi):
        self._readApi = readApi

    def shutdown(self):
        if (
            self._updateNotifyLoopingCall
            and self._updateNotifyLoopingCall.running
        ):
            self._updateNotifyLoopingCall.stop()
            self._updateNotifyLoopingCall = None

        self._process = None
        self._readApi = None
        self._tupleObservable = None

    def _batchNotifyUpdates(self):
        """Batch Notify Updates

        PROBLEM: We don't know if any tuple selectors are already running.
        :return: None
        """
        if not self._updateSelectorQueue:
            return

        num = PeekPsUtil().cpuPercent
        if self.MAX_CPU_PERCENTAGE < num:
            logger.debug("Skipping this loop, CPU is too high: %s", num)
            return

        minDate = min([ts.minDate for ts in self._updateSelectorQueue])
        maxDate = max([ts.maxDate for ts in self._updateSelectorQueue])
        self._updateSelectorQueue = []

        self._tupleObservable.notifyOfTupleUpdate(
            NewEventsTupleSelector(minDate, maxDate)
        )

    @inlineCallbacks
    def importEvents(
        self, modelSetKey: str, eventsEncodedPayload: str
    ) -> Deferred:
        count, maxDate, minDate = yield runPyInPg(
            logger,
            self._dbSessionCreator,
            EventDBImportEventsInPgTask.importEvents,
            None,
            modelSetKey,
            eventsEncodedPayload,
        )

        # Notify anyone watching the events that new ones have arrived.
        if count:
            self._updateSelectorQueue.append(
                NewEventsTupleSelector(minDate, maxDate)
            )

        self._statusController.status.addedEvents += count
        self._statusController.status.lastActivity = datetime.now(pytz.utc)
        self._statusController.notify()

    @inlineCallbacks
    def deleteEvents(self, modelSetKey: str, eventKeys: List[str]) -> Deferred:
        count = yield runPyInPg(
            logger,
            self._dbSessionCreator,
            EventDBImportEventsInPgTask.deleteEvents,
            None,
            modelSetKey,
            eventKeys,
        )

        self._statusController.status.removedEvents += count
        self._statusController.status.lastActivity = datetime.now(pytz.utc)
        self._statusController.notify()

    @inlineCallbacks
    def updateAlarmFlags(
        self, modelSetKey: str, eventKeys: List[str], alarmFlag: bool
    ) -> Deferred:
        count = yield runPyInPg(
            logger,
            self._dbSessionCreator,
            EventDBImportEventsInPgTask.updateAlarmFlags,
            None,
            modelSetKey,
            eventKeys,
            alarmFlag,
        )

        self._statusController.status.updatedAlarmFlags += count
        self._statusController.status.lastActivity = datetime.now(pytz.utc)
        self._statusController.notify()

    @inlineCallbacks
    def replaceProperties(
        self, modelSetKey: str, propertiesEncodedPayload: str
    ) -> Deferred:
        yield runPyInPg(
            logger,
            self._dbSessionCreator,
            EventDBImportPropertiesInPgTask.replaceProperties,
            None,
            modelSetKey,
            propertiesEncodedPayload,
        )

        tupleSelector = TupleSelector(EventDBPropertyTable.tupleName(), {})
        self._tupleObservable.notifyOfTupleUpdate(tupleSelector)

        tupleSelector = TupleSelector(EventDBPropertyTuple.tupleName(), {})
        self._tupleObservable.notifyOfTupleUpdate(tupleSelector)



import logging
from typing import Dict

from peek_plugin_eventdb._private.server.controller.EventDBImportEventsInPgTask import (
    EventDBImportEventsInPgTask,
)
from peek_plugin_eventdb.tuples import loadPublicTuples
from peek_plugin_eventdb.tuples.EventDBPropertyTuple import EventDBPropertyTuple
from vortex.Payload import Payload
from vortex.Tuple import TUPLE_TYPES_BY_NAME

logger = logging.getLogger(__name__)


class EventDBImportPropertiesInPgTask:
    """EventDB Import In PostGreSQL Tasks

    The methods in this class are run in the databases plpython extension.
    """

    @classmethod
    def replaceProperties(
        cls, plpy, modelSetKey: str, propertiesEncodedPayload: bytes
    ) -> None:

        if EventDBPropertyTuple.tupleName() not in TUPLE_TYPES_BY_NAME:
            loadPublicTuples()

        # Reconstruct the data
        properties = Payload().fromEncodedPayload(propertiesEncodedPayload).tuples
        if not properties:
            return

            # Get the model set id
        modelSetId = EventDBImportEventsInPgTask.getModelSetId(
            plpy, modelSetKey, createIfMissing=True
        )

        # Now insert the events
        cls._deleteProperties(plpy, modelSetId)

        # Now insert the events
        cls._insertProperties(plpy, modelSetId, properties)

        # Now insert the events
        cls._insertPropertieValues(plpy, modelSetId, properties)

    @classmethod
    def _insertProperties(cls, plpy, modelSetId: int, properties):

        plan = plpy.prepare(
            """INSERT INTO pl_eventdb."EventDBProperty"
                        ("modelSetId",
                          key, name, "order", comment,
                         "useForFilter", "useForDisplay", "useForPopup", "showFilterAs",
                         "displayByDefaultOnSummaryView",
                         "displayByDefaultOnDetailView")
                        VALUES ($1,
                                $2, $3, $4, $5,
                                $6, $7, $8, $9,
                                $10,
                                $11)""",
            [
                "integer",
                "text",
                "text",
                "integer",
                "text",
                "boolean",
                "boolean",
                "boolean",
                "integer",
                "boolean",
                "boolean",
            ],
        )
        for p in properties:
            plpy.execute(
                plan,
                [
                    modelSetId,
                    p.key,
                    p.name,
                    p.order,
                    p.comment,
                    p.useForFilter,
                    p.useForDisplay,
                    p.useForPopup,
                    p.showFilterAs,
                    p.displayByDefaultOnSummaryView,
                    p.displayByDefaultOnDetailView,
                ],
            )

    @classmethod
    def _insertPropertieValues(cls, plpy, modelSetId: int, properties):
        propIdByKey = cls._loadPropKeyById(plpy, modelSetId)

        plan = plpy.prepare(
            """INSERT INTO pl_eventdb."EventDBPropertyValue"
                            (name, "value", color, comment,
                            "propertyId")
                            VALUES ($1, $2, $3, $4,
                                    $5);""",
            ["text", "text", "text", "text", "integer"],
        )
        for p in properties:
            if p.values:
                for v in p.values:
                    plpy.execute(
                        plan, [v.name, v.value, v.color, v.comment, propIdByKey[p.key]]
                    )

    @classmethod
    def _loadPropKeyById(cls, plpy, modelSetId: int) -> Dict[str, int]:
        sql = """SELECT id, key
                 FROM pl_eventdb."EventDBProperty"
                 WHERE "modelSetId" = %s;"""
        sql %= modelSetId
        rows = plpy.execute(sql, 500)
        return {r["key"]: r["id"] for r in rows}

    @classmethod
    def _deleteProperties(cls, plpy, modelSetId: int):
        sql = """DELETE FROM pl_eventdb."EventDBProperty"
                 WHERE "modelSetId" = %s;"""
        sql %= modelSetId
        plpy.execute(sql)



import logging
from collections import defaultdict
from typing import List, Optional

from peek_plugin_base.storage.LoadPayloadPgUtil import (
    getTuplesPayloadBlocking,
    LoadPayloadTupleResult,
)
from peek_plugin_eventdb._private.server.controller.EventDBController import (
    EventDBController,
)
from peek_plugin_eventdb._private.storage.EventDBModelSetTable import (
    getOrCreateEventDBModelSet,
)
from peek_plugin_eventdb.server.EventDBReadApiABC import EventDBReadApiABC
from peek_plugin_eventdb.tuples.EventDBEventTuple import EventDBEventTuple
from rx.subjects import Subject
from sqlalchemy import select
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class EventDBReadApi(EventDBReadApiABC):
    def __init__(self):
        self._eventdbController = None
        self._dbSessionCreator = None
        self._dbEngine = None

        self._newEventsSubject = defaultdict(Subject)

    def setup(self, eventdbController: EventDBController, dbSessionCreator, dbEngine):
        self._eventdbController = eventdbController
        self._dbSessionCreator = dbSessionCreator
        self._dbEngine = dbEngine

    def shutdown(self):
        pass

    def newEventsObservable(self, modelSetKey: str) -> Subject:
        return self._newEventsSubject[modelSetKey]



"""add useForProp

Peek Plugin Database Migration Script

Revision ID: af33359d5854
Revises: 4ea424ad3883
Create Date: 2020-06-04 21:46:35.963318

"""

# revision identifiers, used by Alembic.
revision = "af33359d5854"
down_revision = "4ea424ad3883"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "EventDBProperty", sa.Column("useForPopup", sa.String()), schema="pl_eventdb"
    )


def downgrade():
    op.drop_column("EventDBProperty", "useForPopup", schema="pl_eventdb")



"""add isAlarm

Peek Plugin Database Migration Script

Revision ID: e24b71e8ebb5
Revises: a63487e16375
Create Date: 2020-07-14 22:24:49.444308

"""

# revision identifiers, used by Alembic.
revision = "e24b71e8ebb5"
down_revision = "a63487e16375"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "EventDBEvent", sa.Column("isAlarm", sa.Boolean()), schema="pl_eventdb"
    )


def downgrade():
    pass
